'''
Copyright 2012 Research Institute eAustria

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Marian Neagul <marian@ieat.ro>
@contact: marian@ieat.ro
@copyright: 2012 Research Institute eAustria
'''

import os
import logging
import json
import threading
import shutil
import time


from mjsrpc2 import RPCBase, jsonmethod, jsonattribute

from me2.server.storage_manager import BtrfsStorageManager
from me2.server.lxc.execution_manager import LxcExecutionManager
from me2.server.bundle_manager import BundleManager

from mako.template import Template
from pkg_resources import ResourceManager

resource_manager = ResourceManager()

class ContainerManager(RPCBase):
    log = logging.getLogger("ContainerManager")
    def __init__(self, context, config):
        RPCBase.__init__(self)
        self.context = context
        self.btrfs = BtrfsStorageManager(context, manager = self, config = config)
        self.lxc = LxcExecutionManager(context, manager = self, config = config)
        self.bundle = BundleManager(context, manager = self, config = config.get("bundles"), storage_manager = self.btrfs)
        self.config = config
        self.package_install_init = Template(text = resource_manager.resource_string(__name__, "data/package_install_init.mak"))
        self.lock = threading.Lock()

    @jsonmethod
    @jsonattribute(name = "bundle_id", kind = str, documentation = "the bundle that should be started")
    @jsonattribute(name = "config", kind = object, documentation = "instance configuration json")
    def start(self, bundle_id, config):
        with self.lock:
            base_bottles = self.btrfs.list_base_bottles()
            if not bundle_id in base_bottles['result']:
                self.bundle.import_bundle(bundle_specification = bundle_id)
            bottle = self.btrfs.clone(base = bundle_id)
            self.log.info("Bottle: %s", bottle)
            container = self.lxc.start(bottle_dir = bottle['path'], config_dir = bottle['configdir'], ident = bottle['uuid'], config = config)
            print container

    @jsonmethod
    @jsonattribute(name = "base_bundle", type = str)
    @jsonattribute(name = "new_bundle", type = str)
    @jsonattribute(name = "packages", type = object)
    @jsonattribute(name = "timeout", kind = int)
    def prepare_bundle(self, base_bundle, new_bundle, packages, timeout):
        base_bottles = self.btrfs.list_base_bottles()
        if not base_bundle in base_bottles['result']:
            self.bundle.import_bundle(bundle_specification = base_bundle)
        bottle = self.btrfs.clone(base = base_bundle)
        init_location = os.path.join(bottle["rootfs"], "sbin", "init")
        init_backup_location = os.path.join(bottle["rootfs"], "sbin", "init.old")
        shutil.move(init_location, init_backup_location)

        if isinstance(packages, basestring):
            packages = packages.split(",")
        content = self.package_install_init.render(packages = packages)
        with open(init_location, "w") as f:
            f.write(content)
        os.chmod(init_location, 755)
        container = self.lxc.start(bottle_dir = bottle['path'], config_dir = bottle['configdir'], ident = bottle['uuid'], config = {})
        start_time = time.time()
        state_file = os.path.join(container.rootfs, 'tmp', 'me2-package-install-state')
        state = None
        self.log.info('Waiting for package install to finish')
        self.log.debug("Timeout is %d", timeout)
        if timeout is None:
            timeout = 600

        while time.time() - start_time < timeout:
            if os.path.exists(state_file):
                state = int(open(state_file).read())
                break
            time.sleep(0.1)

        if state is None:
            self.log.critical("Packages could not be instaled.")
            raise Exception("Could not install packages (maybe an timeout ?)")

        if state == 1:
            self.log.critical("Failed to install packages")
            raise Exception("Package manager failed to install the packages")




class Deployer(RPCBase):
    log = logging.getLogger("Deployer")
    def __init__(self, context, config_file = None):
        RPCBase.__init__(self)
        self.context = context
        if config_file is None:
            self.log.info("No config file provided")
            config = {}
        else:
            self.log.debug("Loading config from %s", config_file)
            config = json.load(open(config_file))
        deployer_config = config.get("deployer", {})
        container_manager_config = deployer_config.get("container-manager", {})
        self.add_method('manager', ContainerManager(context, config = container_manager_config))



import cherrypy
import logging

from mako.template import Template
from mako.lookup import TemplateLookup
from pkg_resources import ResourceManager

resource_manager = ResourceManager()

static_dir = resource_manager.resource_filename(__name__, 'templates/static')
template_dir = resource_manager.resource_filename(__name__, 'templates')

class Root:
    tmpl_lookup = TemplateLookup(directories = [template_dir, ])
    root_tmpl = Template(text = resource_manager.resource_string(__name__, "templates/template.html"), lookup = tmpl_lookup)
    #bundles_tmpl = Template(text = resource_manager.resource_string(__name__, "templates/bundles.html"), lookup = tmpl_lookup)

    def __init__(self, api, api_path):
        self.api = api
        self.api_path = api_path

    @cherrypy.expose
    def index(self):
        return self.render_template('bottles.html')


    @cherrypy.expose
    def js(self):
        cherrypy.response.headers['Content-Type'] = 'application/javascript'
        return ""

    def render_template(self, name, **kwargs):
        tmpl = self.tmpl_lookup.get_template(name)
        args = { 'api': self.api, 'api_path': self.api_path }
        return tmpl.render(**args)




def start_server(address, port, deployer, api_app, api_path = "/api"):
    log = logging.getLogger("start_server")
    log.debug("Static dir: %s", static_dir)
    global_conf = {'global': {'server.socket_host': address,
                       'server.socket_port': int(port),
                       'engine.autoreload_on': False
                       },
            }
    root_conf = {'/static': {
                        'tools.staticdir.on': True,
                        'tools.staticdir.dir': static_dir
                        }}
    cherrypy.tree.mount(Root(api_app, api_path), "/", config = root_conf)
    cherrypy.tree.graft(api_app, '/api')
    cherrypy.config.update(global_conf)
    cherrypy.engine.start()



'''
Copyright 2012 Research Institute eAustria

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Marian Neagul <marian@ieat.ro>
@contact: marian@ieat.ro
@copyright: 2012 Research Institute eAustria
'''

import threading
import logging
import time
from mjsrpc2 import RPCBase, jsonmethod, jsonattribute

from me2.packaging.package import expand_variables

class ContainerSupervisor(threading.Thread, RPCBase):
    log = logging.getLogger("ContainerSupervisor")
    _supervisor_started = False

    def __init__(self):
        threading.Thread.__init__(self)
        RPCBase.__init__(self)
        self.lock = threading.Lock()
        self.map = {}
        self.setDaemon(True)
        self.start()

    def run(self):
        with self.lock:
            if self._supervisor_started:
                self.log.warn("Thread already started")
                return

        self.log.critical("FIXME!!!! Started dummy container supervisor")
        while True:
            time.sleep(1)
            pass


    def add(self, container):
        with self.lock:
            self.log.debug("Registering %s", container.ident)
            if container.ident in self.map:
                raise ValueError("Container already registered!")
            self.map[container.ident] = container
            method_name = "ns_%s" % container.ident
            self.add_method(method_name, container)

    def remove(self, container):
        with self.lock:
            self.log.debug("UnRegistering %s", container.ident)
            return self.map.pop(container.ident)

    def get(self, container_ident):
        with self.lock:
            if container_ident in self.map:
                self.log.error("Container with id=%s is registered!", container_ident)
                return self.map[container_ident]
            else:
                self.log.error("Container with id=%s is NOT registered!", container_ident)
                raise KeyError("Container with id=%s is NOT registered!" % container_ident)

class ExecutionManagerBase(RPCBase):
    log = logging.getLogger("ExecutionManagerBase")
    lock = threading.Lock()
    def __init__(self, context, manager, config):
        super(ExecutionManagerBase, self).__init__()
        self.context = context
        self._manager = manager
        self.config = config
        self._elevate_command = expand_variables("${elevate-privileges}", self.config)

    def _elevate(self, command):
        return self._elevate_command % {'command': command}

    @jsonmethod
    @jsonattribute(name = "bottle_dir", kind = str, documentation = "The bottle directory in which the container should be started")
    @jsonattribute(name = "config_dir", kind = str, documentation = "The config directory for the bottle")
    @jsonattribute(name = "ident", kind = str, documentation = "The identifier of the bottle")
    @jsonattribute(name = "config", kind = object, documentation = "The running config")
    def start(self, bottle_dir, config_dir, ident, config):
        raise NotImplementedError()







import os
from setuptools import find_packages
from distutils.core import setup

def read(fname):
    if os.path.exists(fname):
        return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
      name = "me2-supervisor",
      version = "0.0.8",
      author = "Marian Neagul",
      author_email = "marian@ieat.ro",
      description = "me2 is a mOSAIC Cloud Project component responsible for the deployment and supervision of application",
      license = "APL",
      keywords = "deployment mosaic",
      url = "http://developers.mosaic-cloud.eu",
      namespace_packages = ["me2"],
      package_dir = {'':'src'},
      packages = find_packages("src", exclude = ["*tests*", ]),
      long_description = read('README.rst'),
      include_package_data = True,
      package_data = {
                      'me2.server.lxc': ["data/*.mak", "data/templates/*.mak"],
                      'me2.server.web': ["templates/static/*/*", "templates/*.html"],
                      'me2.server': ["data/*.mak"]
                      },
      classifiers = [
                     "Intended Audience :: Developers",
                     "Development Status :: 3 - Alpha",
                     "Topic :: System :: Installation/Setup",
                     "License :: OSI Approved :: Apache Software License",
                     ],
      entry_points = {
                      'console_scripts': [
                                          'mded = me2.server.apps.mded:main',
                                          ]
                      },
      dependency_links = [
        "http://developers.mosaic-cloud.eu/artifactory/pypi/"
        ],
      install_requires = ["mjsrpc2==0.0.6", "lockfile>=0.9.0", "python-daemon>=1.6", "pyyaml>=3.0", "mako>=0.6.0", "me2-packager==0.0.5", "CherryPy>=3.0.0"]
)



'''
Copyright 2012 Research Institute eAustria

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Marian Neagul <marian@ieat.ro>
@contact: marian@ieat.ro
@copyright: 2012 Research Institute eAustria
'''

import argparse
import logging
import daemon
import os

class PidFileException(Exception):
    pass

class PidfileManager(object):
    def __init__(self, fname):
        self.fname = fname
    def __enter__(self):
        if os.path.exists(self.fname):
            raise PidFileException("Pid file exists")
        self._pidfile = open(self.fname, 'w')
        self._pidfile.write(str(os.getpid()))
        self._pidfile.flush()
    def __exit__(self, type, value, traceback):
        self._pidfile.close()
        os.remove(self.fname)


def main():
    parser = argparse.ArgumentParser(description = 'mOSAIC deployment daemon')
    parser.add_argument('--stdout', '-o', type = str, default = None, help = 'File for storing stdout')
    parser.add_argument('--stderr', '-e', type = str, default = None, help = 'File for storing stderr')
    parser.add_argument('--verbose', '-v', action = 'count', default = 0, help = "The verbosity level. More v's means more logging")
    daemon_group = parser.add_argument_group('Daemon')
    daemon_group.add_argument('--chroot', '-c', type = str, default = None, help = 'Directory where we should chroot')
    daemon_group.add_argument('--workdir', '-w', type = str, default = "/", help = 'Working directory')
    daemon_group.add_argument('--pidfile', '-f', type = str, default = None, help = 'PID File')
    daemon_group.add_argument('--detach', '-d', default = False, action = 'store_true', help = 'Daemonize')
    daemon_group.add_argument('--logfile', '-l', type = str, default = None, help = 'File for storing log output (defaults to --stderr)')
    daemon_group.add_argument('--config-file', type = str, default = None, help = 'Configuration file (JSON)')

    app_group = parser.add_argument_group('Application')
    unix_group = parser.add_argument_group('Unix Domain Socket Server')
    tcp_group = parser.add_argument_group('TCP Server')
    agent_group = parser.add_argument_group('Agent')

    app_group.add_argument("--application", "-a", type = str, default = [], action = 'append', help = 'Enable an internal application')
    #app_group.add_argument("--noapi", default = False, action = 'store_true', help = 'Disable API server')
    app_group.add_argument("--enable-unix-server", '-u', action = 'store_true', default = False, help = 'Enable the unix socket server')
    unix_group.add_argument("--unix-socket", '-s', type = str, default = os.path.join(os.path.expanduser("~"), '.mde.socket'), help = 'The UNIX socket that should be used')
    app_group.add_argument("--enable-tcp-server", '-t', action = 'store_true', default = False, help = 'Enable TCP server')
    tcp_group.add_argument("--bind", '-b', type = str, default = '127.0.0.1', help = 'The address to bind to')
    tcp_group.add_argument("--port", '-p', type = str, default = '8099', help = 'The port to bind to')

    agent_group.add_argument('--agent-bundle', type = str, default = None, help = 'The agent bundle')


    args = parser.parse_args()

    # Setup logging
    if args.verbose > 4:
        log_level = 10
    else:
        log_level = 50 - args.verbose * 10

    logging.basicConfig(format = '[ %(levelname)-8s %(filename)s:%(lineno)d (%(name)s) ] --> %(message)s',
                        datefmt = '%d/%b/%Y %H:%M:%S',
                        level = log_level)


    from me2.server import daemon as mde_daemon

    context = daemon.DaemonContext()
    context.chroot_directory = args.chroot
    context.working_directory = args.workdir

    if args.pidfile is not None:
        context.pidfile = PidfileManager(args.pidfile)
    context.detach_process = args.detach

    if args.stdout is not None:
        context.stdout = open(args.stdout, 'w+')

    if args.stderr is not None:
        context.stderr = open(args.stderr, 'w+')
    elif context.stdout is not None:
        context.stderr = open(args.stdout, 'w+')


    context.application = args.application
    #context.noapi = args.noapi
    context.enable_unix_server = args.enable_unix_server
    context.unix_socket = args.unix_socket
    context.enable_tcp_server = args.enable_tcp_server
    context.tcp_bind = args.bind
    context.tcp_port = args.port
    if args.config_file:
        context.config_file = os.path.abspath(args.config_file)
    else:
        context.config_file = None

    context.agent_bundle = args.agent_bundle


    with context:
        if args.logfile is None:
            if args.stderr is not None:
                context.logfile = args.stderr
            elif args.stdout is not None:
                context.logfile = args.stdout
            else:
                context.logfile = "/dev/null" # Platform specific
        else:
            context.logfile = args.logfile

        logging.basicConfig(level = logging.DEBUG,
                    format = '%(asctime)s %(threadName)-10s %(name)-16s %(levelname)-8s %(message)s',
                    datefmt = '%m-%d %H:%M',
                    filename = context.logfile,
                    filemode = 'w+'
                    )
        mde_daemon.main(context)

if __name__ == "__main__":
    import sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../")))
    main()
